#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <qdos.h>
#include <qptr.h>
#include "csrvthg.h"

short __returnCSP__ = 0;

typedef struct
{
    short   len;
    char    txt[1];
} CSTEXT;

typedef struct 
{
    long cookie1;
    short plen;
    char pnam[22];
    long cookie2;
    short len;
    char txt[1];
} MAGIC;

typedef struct
{
    long    bchn;
    short   len;
    char    txt[1];
}  CSREQST;

typedef struct
{
    long  chn;
    short len;
} CSPOLL;


typedef struct
{
    int (*creq)(CSTEXT *, CSREQST *, short *);
    int (*cpoll)(CSTEXT *, short, CSPOLL *);
    int (*lpoll)(CSTEXT *, short, CSPOLL *);
    int (*sbro)(CSTEXT *, CSTEXT *);
    unsigned long vers;
    int (*spend)(CSTEXT *);
} __CSPROCS__;


extern int CSUse (int, void *, int *);
extern void CSFree (int, CSTEXT *);

static __CSPROCS__ *csp = NULL;
static short ispend = 0;

#define COOKIE1 0xdeadbeef;
#define COOKIE2 0xfeedface;

static CSTEXT * SetText(char *s)
{
    short n;
    CSTEXT *p;
    p =  malloc((n = strlen(s)) + sizeof(CSTEXT));
    memcpy (p->txt, s, n);
    p->len = n;
    return p;
}

int UseSrvThgL (int flag, char *name, int plen)
{
    MAGIC *p;
    int rv = -1;
    int tsp;
    int n;

    p = calloc(1, ((n = strlen(name))+sizeof(MAGIC)));

    if(p)
    {
        if(plen)
        {
            p->cookie1 = COOKIE1;
            strcpy(p->pnam, "pipe_");
            itoa(plen, p->pnam+5);
            p->plen = strlen(p->pnam);
            p->cookie2 = COOKIE2;
        }
        memcpy (p->txt, name, n);
        p->len = n;

        rv = CSUse(flag ,&(p->len), &tsp);
        if(rv > 0)
        {
            if(csp == NULL)
            {
                csp = (__CSPROCS__ *) rv;
            }
            if(csp->vers > SPVERS)
            {
                ispend = 1;
            }
            if(csp->vers < CVERS)
            {
                rv = -19;
            }
            else if(!__returnCSP__)
            {
                rv = 0;
            }
            else
            {
                rv = tsp;
            }
        }
        free(p);
    }
    return rv;
}

int UseSrvThg(int flag, char *name)
{
    return UseSrvThgL(flag, name, 0);
}

void FreeSrvThg(int flag, char *name)
{
    CSTEXT *p;
    if( p = SetText(name))
    {
        CSFree (flag, p);
        free(p);
    }
}

void * Request (char *name, void *cmd, short len, short *rlen)
{
    char *buf = NULL;

    if(csp)
    {
        CSTEXT *pn;
        CSREQST *pcmd;
        if(pn = SetText(name))
        {
            if(len == -1)
            {
                len = strlen(cmd);
            }
            if(pcmd = malloc(len + sizeof(CSREQST)))
            {
                int rchn;

                pcmd->bchn = 0;
                pcmd->len = len;
                memcpy(pcmd->txt, cmd,len);
                rchn = (csp->creq)(pn, pcmd, rlen);

                if(rchn > 0 && *rlen > 0)
                {
                    if(buf = malloc ((*rlen)+1))
                    {
                        int res;
                        res = io_fstrg(rchn, -1, buf, *rlen);
                        if(res < 0)
                        {
                            free(buf);
                            buf = NULL;
                            *rlen = 0;
                        }
                        else
                        {
                            *(buf + *rlen) = 0;
                        }
                    }
                }
                free(pcmd);
            }
            free(pn);
	}            
    }
    return buf;        
}


static int dopoll (char *name, short timeout, void **pcmd, short *rlen,
    int (*func)(CSTEXT *, short, CSPOLL *))
{
    CSTEXT *pn;
    int rchn = -3;

    if(csp)
    {
	if(pn = SetText(name))
	{
	    CSPOLL poll;
	    poll.chn = poll.len = 0;
	    rchn = (func)(pn, timeout, &poll);
	    if(rchn > 0)
	    {
		if(poll.len > 0 && poll.chn >= 0)
		{
		    if(*pcmd = malloc(poll.len+1))
		    {
			int tchn = io_fstrg(rchn, -1, (*pcmd), poll.len);
			if(tchn < 0)
			{
			    rchn = tchn;
			    free(*pcmd);
			    *pcmd = NULL;
			    poll.len = 0;
			}
			else
			{
			    char *pp = *pcmd;
			    rchn = poll.chn;
			    *(pp + poll.len) = 0;
			}
		    }
		    else
		    {
			rchn = -3;
		    }
		}
	    }        
	    *rlen = poll.len;
	    free(pn);
	}
    }
    return rchn;
}

int SrvPoll (char *name, short timeout, void **pcmd, short *rlen)
{
    return dopoll (name, timeout, pcmd, rlen, csp->cpoll);
}
int LstnPoll (char *name, short timeout, void **pcmd, short *rlen)
{
    return dopoll (name, timeout, pcmd, rlen, csp->lpoll);
}

int SrvRply (int rchn, void *reply, short len)
{
    int res = 0;
    if(rchn)
    {
    if(reply == NULL)
    {
        len = 0;
    } 
    else if(len == -1)
    {
        len = strlen(reply);
    }
    if((res = io_sstrg(rchn, -1, &len, 2)) > 0)
    {
        if(len)
        {
            res = io_sstrg(rchn, -1, reply, len);
        }
    }
    }
    return res;
}

int SrvBro (char *name, void *cmd, short len)
{
    int rchn = -1;
    if(csp)
    {
        CSTEXT *pn;
        CSTEXT *pcmd;
        if(pn = SetText(name))
        {
            if(len == -1)
            {
                len = strlen(cmd);
            }
            if(pcmd = malloc(len + sizeof(CSTEXT)))
            {
                pcmd->len = len;
                memcpy(pcmd->txt, cmd,len);
                rchn = (csp->sbro)(pn, pcmd);
                free(pcmd);
	    }
            free(pn);
	}
    }
    return rchn;
}

int SrvPend (char *name)
{
    int rchn = -1;
    if(ispend && csp)
    {
        CSTEXT *pn;
        CSTEXT *pcmd;
        if(pn = SetText(name))
        {
            rchn = (csp->spend)(pn);
	}
    }
    return rchn;
}
